/* cvt.c

Convert numbers from one base to another.  Input and output bases are
specified with "-i<base>" and "-o<base>", where base can be any of
b,d,o,u,x (binary, decimal, octal, unsigned decimal, hex).  The number
to be converted can be a command-line argument or come from stdin;
output is to stdout. - r l st.peters, 12/7/88

19950926.  "Base" can also now be 'n' for network IP addresses - not
really a base but a different format.

/*
 * Copyright (c) 1988,1995,1999, Richard L. St.Peters
 * You may use this software for any legal purpose, including giving it
 * to others, and/or modifying it, provided this copyright and disclaimer
 * remains intact.
 *
 * This software has no warranty whatsoever.  It is useless junk.  */

#include <stdio.h>

#define		equals(s1,s2)		(!strcmp((s1),(s2)))

extern char *optarg ;
extern int optind, opterr;
char *options = "hi:o:" ;
char *itype = NULL, *otype = NULL, *ibuf = NULL ;
char *prog ;
char *basename (char *path) ;

main (int argc, char **argv)
{
    unsigned int value ;
    int i ;
    
    prog = basename (argv[0]) ;
    while ((i = getopt (argc, argv, options)) != EOF)
    switch (i)
    {
      case 'i':
	itype = optarg ;
	break ;
      case 'o':
	otype = optarg ;
	break ;
      case 'h':
	usage (prog) ;
	break ;
    }

    if (otype == NULL)
      otype="u" ;

    if (itype == NULL)
      itype = "u" ;

    if (argc > optind)
      ibuf = argv[optind] ;
    read_input (&value) ;
    print_output (value) ;
}

read_input (unsigned int p)
{
    char format[BUFSIZ], buf[BUFSIZ] ;

    if (ibuf == NULL) 
    {
	ibuf = buf ;
	gets (ibuf) ;
    }

    if (*itype == 'b')
      read_binary (p) ;
    else if (*itype == 'n')
      read_ip (p) ;
    else 
    {
	sprintf (format, "%%%c", *itype) ;
	sscanf (ibuf, format, p) ;
    }
}

print_output (int n)
{
    char format[BUFSIZ] ;

    if (*otype == 'b')
      print_binary (n) ;
    else if (*otype == 'n')
      print_ip (n) ;
    else
    {
	sprintf (format, "%%%c", *otype) ;
	printf (format, n) ;
    }
    if (n < 256 && *otype != 'n' && *itype != 'n')
      print_ascii (n) ;

    printf ("\n") ;
}

read_binary (unsigned int *p)
{
    int i, N, M ;

    *p = 0 ;
    N = strlen (ibuf) ;
    M = N-1 ;
    for (i = 0 ; i < N ; i++)
      *p += ibuf[i] == '1' ? 1 << M-i : 0 ;
}

read_ip (unsigned int *p)
{
    unsigned int i, f[4] ;
    char *q, *q0 = ibuf ;

    q = q0;
    for (i=0; i<4; i++) 
    {
	while (*q != '.' && *q != '\0') q++ ;
	*q = '\0' ;
	sscanf (q0, "%d", &f[3-i]) ;
	q0 = ++q ;
    }
    *p = (f[3]<<24) | (f[2]<<16) | (f[1]<<8) | (f[0]) ;
}

print_ip (unsigned int n)
{
    printf ("%u.%u.%u.%u", (n>>24)&256-1, (n>>16)&256-1,
	    (n>>8)&256-1, n&256 - 1) ;
}

print_binary (int n)
{
    int i, N = 8 * sizeof (int) ;
    char buf[BUFSIZ] ;

    buf[N] = '\0' ;
    for (i = 1 ; i <= N; i++, n >>= 1)
      buf[N-i] = (n & 1 ? '1' : '0') ;

    printf ("%s", buf) ;
}

char *basename (char *p)
{
    char *q = p ;

    while (*q != '\0')
      if (*q == '/' && q[1] != '\0')
	p = ++q ;
      else
	q++ ;

    return (p) ;
}

print_ascii (int input)
{
    printf (", ascii = ") ;

    if (input & 0200)
    {
	printf ("meta-") ;
	input &= 0177 ;
    }

    if (input == '\177')
      printf ("^%c", '?') ;
    else if (input == 0)
      printf ("^%c", '@') ;
    else if (input < 27)
      printf ("^%c", input + 'A' - 1) ;
    else if (input == ' ')
      printf ("space") ;
    else if (input > ' ')
      printf ("%c", input) ;
    else
      switch (input)
      {
	case '\033':
	  printf ("^%c", '[') ;
	  break ;
	case '\034':
	  printf ("^%c", '\\') ;
	  break;
	case '\035':
	  printf ("^%c", ']') ;
	  break ;
	case '\036':
	  printf ("^%c", '~') ;
	  break ;
	case '\037':
	  printf ("^%c", '/') ;
	  break ;
      }
}

usage (char *p)
{
    printf ("Usage:  %s [-i<input type>] -o<output type> [number]\n", p) ;
    printf ("Available types are b,d,n,o,u,x") ;
    printf (" (binary, decimal, network, octal, unsigned decimal, hex)\n") ;
    printf ("Input and output types default to u if unspecified.\n") ;
    exit () ;
}

